/*
 * IRC - Internet Relay Chat, ircd/querycmds.c (formerly ircd/s_serv.c)
 * Copyright (C) 1990 Jarkko Oikarinen and
 *		      University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef __GNUC__
#pragma implementation
#endif
#include "sys.h"
#include <sys/stat.h>
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <stdlib.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "h.h"
#include "struct.h"
#include "parse.h"
#include "send.h"
#include "s_err.h"
#include "numeric.h"
#include "ircd.h"
#include "s_user.h"
#include "version.h"
#include "s_bsd.h"
#include "s_misc.h"
#include "match.h"
#include "s_serv.h"
#include "msg.h"
#include "channel.h"
#include "numnicks.h"
#include "userload.h"
#include "s_conf.h"
#include "support.h"

RCSTAG_CC("$Id: querycmds.c,v 1.3 1997/09/12 02:21:06 carlo Exp $")

/*
 * m_functions execute protocol messages on this server:
 *
 *   cptr    is always NON-NULL, pointing to a *LOCAL* client
 *	     structure (with an open socket connected!). This
 *	     identifies the physical socket where the message
 *	     originated (or which caused the m_function to be
 *	     executed--some m_functions may call others...).
 *
 *   sptr    is the source of the message, defined by the
 *	     prefix part of the message if present. If not
 *	     or prefix not found, then sptr==cptr.
 *
 *	     (!IsServer(cptr)) => (cptr == sptr), because
 *	     prefixes are taken *only* from servers...
 *
 *	     (IsServer(cptr))
 *		     (sptr == cptr) => the message didn't
 *		     have the prefix.
 *
 *		     (sptr != cptr && IsServer(sptr) means
 *		     the prefix specified servername. (?)
 *
 *		     (sptr != cptr && !IsServer(sptr) means
 *		     that message originated from a remote
 *		     user (not local).
 *
 *	     combining
 *
 *	     (!IsServer(sptr)) means that, sptr can safely
 *	     taken as defining the target structure of the
 *	     message in this server.
 *
 *   *Always* true (if 'parse' and others are working correct):
 *
 *   1)	     sptr->from == cptr	 (note: cptr->from == cptr)
 *
 *   2)	     MyConnect(sptr) <=> sptr == cptr (e.g. sptr
 *	     *cannot* be a local connection, unless it's
 *	     actually cptr!). [MyConnect(x) should probably
 *	     be defined as (x == x->from) --msa ]
 *
 *   parc    number of variable parameter strings (if zero,
 *	     parv is allowed to be NULL)
 *
 *   parv    a NULL terminated list of parameter pointers,
 *
 *		     parv[0], sender (prefix string), if not present
 *			     this points to an empty string.
 *		     parv[1]...parv[parc-1]
 *			     pointers to additional parameters
 *		     parv[parc] == NULL, *always*
 *
 *	     note:   it is guaranteed that parv[0]..parv[parc-1] are all
 *		     non-NULL pointers.
 */

/*
 * m_version
 *
 *   parv[0] = sender prefix
 *   parv[1] = remote server
 */
int m_version(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  Reg1 aClient *acptr;

  if (MyConnect(sptr) && parc > 1)
  {
    if (!(acptr = find_match_server(parv[1])))
    {
      sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], parv[1]);
      return 0;
    }
    parv[1] = acptr->name;
  }

  if (hunt_server(0, cptr, sptr, ":%s VERSION :%s", 1, parc, parv) ==
      HUNTED_ISME)
    sendto_one(sptr, rpl_str(RPL_VERSION),
        me.name, parv[0], version, debugmode, me.name, serveropts);

  return 0;
}

/*
 * m_info
 *
 * parv[0] = sender prefix
 * parv[1] = servername
 */
int m_info(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  char **text = infotext;

  if (hunt_server(1, cptr, sptr, ":%s INFO :%s", 1, parc, parv) == HUNTED_ISME)
  {
    while (text[2])
    {
      if (!IsOper(sptr))
	sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], *text);
      text++;
    }
    if (IsOper(sptr))
    {
      while (*text)
	sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], *text++);
      sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], "");
    }
    sendto_one(sptr, ":%s %d %s :Birth Date: %s, compile # %s",
	me.name, RPL_INFO, parv[0], creation, generation);
    sendto_one(sptr, ":%s %d %s :On-line since %s",
	me.name, RPL_INFO, parv[0], myctime(me.firsttime));
    sendto_one(sptr, rpl_str(RPL_ENDOFINFO), me.name, parv[0]);
  }
  return 0;
}

/*
 * m_links
 *
 * parv[0] = sender prefix
 * parv[1] = servername mask
 *
 * or
 *
 * parv[0] = sender prefix
 * parv[1] = server to query
 * parv[2] = servername mask
 */
int m_links(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  char *mask;
  aClient *acptr;

  if (parc > 2)
  {
    if (hunt_server(1, cptr, sptr, ":%s LINKS %s :%s", 1, parc, parv) !=
        HUNTED_ISME)
      return 0;
    mask = parv[2];
  }
  else
    mask = parc < 2 ? NULL : parv[1];

  for (acptr = client, collapse(mask); acptr; acptr = acptr->next)
  {
    if (!IsServer(acptr) && !IsMe(acptr))
      continue;
    if (!BadPtr(mask) && match(mask, acptr->name))
      continue;
    sendto_one(sptr, rpl_str(RPL_LINKS),
	me.name, parv[0], acptr->name, acptr->serv->up->name,
#ifndef GODMODE
	acptr->hopcount, acptr->serv->prot,
#else /* GODMODE */
	acptr->hopcount, acptr->serv->prot, acptr->serv->timestamp,
	NumServ(acptr),
#endif /* GODMODE */
	(acptr->info[0] ? acptr->info : "(Unknown Location)"));
  }

  sendto_one(sptr, rpl_str(RPL_ENDOFLINKS), me.name, parv[0],
      BadPtr(mask) ? "*" : mask);
  return 0;
}

/*
 * m_help
 *
 * parv[0] = sender prefix
 */
int m_help(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  int i;

  for (i = 0; msgtab[i].cmd; i++)
    sendto_one(sptr, ":%s NOTICE %s :%s", me.name, parv[0], msgtab[i].cmd);
  return 0;
}

/*
 * m_lusers
 *
 * parv[0] = sender
 * parv[1] = host/server mask.
 * parv[2] = server to query
 */
int m_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  int s_count = 0, c_count = 0, u_count = 0, i_count = 0;
  int o_count = 0, m_client = 0, m_client_local = 0, m_server = 0;
  char mydom_mask[HOSTLEN + 1];
  aClient *acptr;

  if (parc > 2)
    if (hunt_server(1, cptr, sptr, ":%s LUSERS %s :%s", 2, parc, parv) !=
        HUNTED_ISME)
      return 0;

  mydom_mask[0] = '*';
  strncpy(&mydom_mask[1], DOMAINNAME, HOSTLEN - 1);
  mydom_mask[HOSTLEN] = 0;

  collapse(parv[1]);
  for (acptr = client; acptr; acptr = acptr->next)
  {
    if (parc > 1)
      if (!IsServer(acptr) && acptr->user)
      {
	if (match(parv[1], acptr->user->server->name))
	  continue;
      }
      else if (match(parv[1], acptr->name))
	continue;

    switch (acptr->status)
    {
      case STAT_SERVER:
	if (MyConnect(acptr))
	  m_server++;
      case STAT_ME:
	s_count++;
	break;
      case STAT_CLIENT:
	if (IsOper(acptr))
	  o_count++;
	if (MyConnect(acptr))
	{
	  m_client++;
	  if (match(mydom_mask, acptr->sockhost) == 0)
	    m_client_local++;
	}
	if (!IsInvisible(acptr))
	  c_count++;
	else
	  i_count++;
	break;
      case STAT_PING:
	break;
      default:
	u_count++;
	break;
    }
  }
  sendto_one(sptr, rpl_str(RPL_LUSERCLIENT),
      me.name, parv[0], c_count, i_count, s_count);
  if (o_count)
    sendto_one(sptr, rpl_str(RPL_LUSEROP), me.name, parv[0], o_count);
  if (u_count > 0)
    sendto_one(sptr, rpl_str(RPL_LUSERUNKNOWN), me.name, parv[0], u_count);
  if ((c_count = count_channels(sptr)) > 0)
    sendto_one(sptr, rpl_str(RPL_LUSERCHANNELS),
        me.name, parv[0], count_channels(sptr));
  sendto_one(sptr, rpl_str(RPL_LUSERME), me.name, parv[0], m_client, m_server);
  if (MyClient(sptr) || Protocol(cptr) < 10)
    sendto_one(sptr,
	":%s NOTICE %s :Highest connection count: %d (%d clients)",
	me.name, parv[0], max_connection_count, max_client_count);
  else
    sendto_one(sptr,
	"%c NOTICE %c%c%c :Highest connection count: %d (%d clients)",
	NumServ(&me), NumNick(sptr), max_connection_count, max_client_count);
  if (m_client > max_client_count)
    max_client_count = m_client;
  if ((m_client + m_server) > max_connection_count)
  {
    max_connection_count = m_client + m_server;
    if (max_connection_count % 10 == 0)		/* only send on even tens */
      sendto_ops("Maximum connections: %d (%d clients)",
	  max_connection_count, max_client_count);
  }
  current_load.local_count = m_client_local;
  current_load.client_count = m_client;
  current_load.conn_count = m_client + m_server;
  return 0;
}

/*
 * m_admin
 *
 * parv[0] = sender prefix
 * parv[1] = servername
 */
int m_admin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  aConfItem *aconf;

  if (MyConnect(sptr) && parc > 1)
  {
    aClient *acptr;
    if (!(acptr = find_match_server(parv[1])))
    {
      sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], parv[1]);
      return 0;
    }
    parv[1] = acptr->name;
  }
  if (hunt_server(0, cptr, sptr, ":%s ADMIN :%s", 1, parc, parv) != HUNTED_ISME)
    return 0;
  if ((aconf = find_admin()))
  {
    sendto_one(sptr, rpl_str(RPL_ADMINME), me.name, parv[0], me.name);
    sendto_one(sptr, rpl_str(RPL_ADMINLOC1), me.name, parv[0], aconf->host);
    sendto_one(sptr, rpl_str(RPL_ADMINLOC2), me.name, parv[0], aconf->passwd);
    sendto_one(sptr, rpl_str(RPL_ADMINEMAIL), me.name, parv[0], aconf->name);
  }
  else
    sendto_one(sptr, err_str(ERR_NOADMININFO),
	me.name, parv[0], me.name);
  return 0;
}

/*
 * m_motd
 *
 * parv[0] - sender prefix
 * parv[1] - servername
 *
 * modified 30 mar 1995 by flux (cmlambertus@ucdavis.edu)
 * T line patch - display motd based on hostmask
 */
int m_motd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
  int fd, nr;
  char line[80];
  Reg1 char *tmp;
  struct tm *tm;
  aConfItem *ftmp;
  int is_remote = 0;
  int count;

  if (hunt_server(0, cptr, sptr, ":%s MOTD :%s", 1, parc, parv) != HUNTED_ISME)
    return 0;

  /*
   * Find out if this is a remote query or if we have a T line for our hostname
   */
  if (IsServer(cptr))
    is_remote = 1;
  else
    ftmp = find_tline(cptr);

  /*
   * Stop NFS hangs...most systems should be able to open a file in
   * 3 seconds. -avalon (curtesy of wumpus)
   */
  alarm(3);
  /* If we returned 0 in find_tline, there is no T line for this host
   * so display the stock ircd.motd
   * If its a remote query, use the remote.motd	 -Ghostwolf
   */
  if (is_remote)
    fd = open(RMOTD, O_RDONLY);
  else if (!ftmp || (fd = open(ftmp->passwd, O_RDONLY)) == -1)
    fd = open(MOTD, O_RDONLY);
  alarm(0);
  if (fd == -1)
  {
    sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0]);
    return 0;
  }
  if (!is_remote)
  {
    struct stat sb;
    fstat(fd, &sb);
    sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name);
    tm = localtime((time_t *) & sb.st_mtime);	/* NetBSD needs cast */
    sendto_one(sptr, ":%s %d %s :- %d/%d/%d %d:%02d", me.name, RPL_MOTD,
	parv[0], tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year,
	tm->tm_hour, tm->tm_min);
    count = 100;
  }
  else
    count = 3;
  dgets(-1, NULL, 0);		/* Make sure buffer is at empty pos */
  while ((nr = dgets(fd, line, sizeof(line) - 1)) > 0)
  {
    line[nr] = '\0';
    if ((tmp = strchr(line, '\n')))
      *tmp = '\0';
    if ((tmp = strchr(line, '\r')))
      *tmp = '\0';
    sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], line);
    if (--count == 0)
      break;
  }
  dgets(-1, NULL, 0);		/* Make sure buffer is at empty pos */
  sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0]);
  close(fd);
  return 0;
}
